/** @file cfl_graphfncts.cpp Operations on (directed) graphs. */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2009  Thomas Moor, Klaus Schmidt, Sebastian Perk
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "cfl_graphfncts.h"


namespace faudes {

/*
**********************************************************************
**********************************************************************
**********************************************************************

Search for strongly connected ycless components (SCC)

**********************************************************************
**********************************************************************
**********************************************************************
*/

// Filter class for SCC search: statics
const StateSet SccFilter::msEmptyStates=StateSet();
const EventSet SccFilter::msEmptyEvents=EventSet();

// Filter class for SCC search: destructor
SccFilter::~SccFilter(void) {
  if(mpStatesAvoid) delete mpStatesAvoid;
  if(mpStatesRequire) delete mpStatesRequire;
  if(mpEventsAvoid) delete mpEventsAvoid;
}

// Filter class for SCC search: constructors
SccFilter::SccFilter(void) :
  mMode(0x00),
  pStatesAvoid(&msEmptyStates),
  pStatesRequire(&msEmptyStates),
  pEventsAvoid(&msEmptyEvents),
  mpStatesAvoid(0),
  mpStatesRequire(0),
  mpEventsAvoid(0)
{}

// Filter class for SCC search: constructors
SccFilter::SccFilter(int mode, const StateSet& rStatesAvoidRequire) :
  mMode(mode),
  pStatesAvoid(&msEmptyStates),
  pStatesRequire(&msEmptyStates),
  pEventsAvoid(&msEmptyEvents),
  mpStatesAvoid(0),
  mpStatesRequire(0),
  mpEventsAvoid(0)
{
  // no special events
  mMode &= ~EventsAvoid;
  // use states for require
  if(mMode & StatesRequire)
    pStatesRequire=&rStatesAvoidRequire;
  // use states for avoid
  if(mMode & StatesAvoid)
    pStatesAvoid=&rStatesAvoidRequire;
}

// Filter class for SCC search: constructors
SccFilter::SccFilter(int mode, const StateSet& rStatesAvoid, const StateSet& rStatesRequire): 
  mMode(mode),
  pStatesAvoid(&msEmptyStates),
  pStatesRequire(&msEmptyStates),
  pEventsAvoid(&msEmptyEvents),
  mpStatesAvoid(0),
  mpStatesRequire(0),
  mpEventsAvoid(0)
{
  // no special events
  mMode &= ~EventsAvoid;
  // but special states
  mMode |= StatesRequire;
  mMode |= StatesAvoid;
  // set internal pointers
  pStatesAvoid=&rStatesAvoid;
  pStatesRequire=&rStatesRequire;
}


// Filter class for SCC search: constructors
SccFilter::SccFilter(int mode, const Generator& rGen) :
  mMode(mode),
  pStatesAvoid(&msEmptyStates),
  pStatesRequire(&msEmptyStates),
  pEventsAvoid(&msEmptyEvents),
  mpStatesAvoid(0),
  mpStatesRequire(0),
  mpEventsAvoid(0)
{
  // sets all empty by default
  mMode &= ~(StatesAvoid | StatesRequire | EventsAvoid );
  // ignore livelocks ==> require at least one marked state
  if(mode & IgnoreLiveLocks) {
    pStatesRequire=&rGen.MarkedStates();
    mMode |= StatesRequire;
  }
  // livelocks only ==> avoid marked states
  if(mode & LiveLocksOnly) {
    pStatesAvoid=&rGen.MarkedStates();
    mMode |= StatesAvoid;
  }
  // accessible only ==> deal with this in api wrappers
  if(mode & IgnoreUnaccessible) {
    (void) mode;
  }  
}



// Recursive function to search for SCCs.
// The algorithm is a minor variation of the one prsented in
//  -- Aho, Hopcroft, Ullman: The Design and Analysis of Computer Algorithms
void SearchScc(
  const Idx vState,            // Current state under investigation
  int& vRcount,                // Depth count of recursion
  const Generator& rGen,       // Graph to inspect
  const SccFilter& rFilter,    // Filter to ignore transitions with specified events
  StateSet&  rTodo,            // States not dealt with so far
  std::stack<Idx>& rStack,      // Stack of currently considered states
  StateSet& rStackStates,      // Stack contents as set
  std::map<const Idx, int>& rDfn, // Map assigning to each state its depth-first count
  std::map<const Idx, int>& rLowLnk, // Map assigning to each state its LOWLINK Number
  std::list<StateSet>& rSccList,  // Record result: list of SCCs 
  StateSet& rRoots)             // Record result: one state per SCC
{
  FD_DF("SerchScc: -- recursive search from state "<< vState << " at level " << vRcount << " --");
  // Take current state from todo list
  rTodo.Erase(vState);
  // Ignore filter: avoid states
  if(rFilter.mMode & SccFilter::StatesAvoid)
  if(rFilter.pStatesAvoid->Exists(vState)) return;
  // Ignore filter: break recursion on find first
  if(rFilter.mMode & SccFilter::FindFirst)
  if(!rSccList.empty()) return;
  // Record depth-first level for current state and initialise low-link level
  rDfn[vState]=vRcount;
  rLowLnk[vState]=vRcount;
  // Increment depth-first level
  vRcount++;
  // Push state on stack;
  rStack.push(vState);
  rStackStates.Insert(vState);
  /*
  std::list<Idx>::iterator vsit = --rStack.end();
  */
  // Create set of successor states "L[state]" 
  StateSet SuccStates = StateSet();
  TransSet::Iterator tit = rGen.TransRelBegin(vState);
  TransSet::Iterator tit_end = rGen.TransRelEnd(vState);
  for(; tit != tit_end; ++tit) {
    // Ignore avoid: avoid states
    if(rFilter.mMode & SccFilter::StatesAvoid)
    if(rFilter.pStatesAvoid->Exists(tit->X2)) continue;
    // Ignore filter: avoid events
    if(rFilter.mMode & SccFilter::EventsAvoid)
    if(rFilter.pEventsAvoid->Exists(tit->Ev)) continue;
    // Record
    SuccStates.Insert(tit->X2);
  }
  // Iterate over vertices *sit of L[state] 
  StateSet::Iterator sit = SuccStates.Begin();
  StateSet::Iterator sit_end = SuccStates.End();
  for(; sit != sit_end; ++sit) {
    // Successors that are on the todo list get ... 
    if(rTodo.Exists(*sit)) {        
       // ... searched recursively, and ...
       SearchScc(*sit, vRcount, rGen, rFilter, rTodo, rStack, rStackStates, rDfn, rLowLnk, rSccList, rRoots);
       // ... the current low-link gets updated to the new minimum.
       if(rLowLnk[*sit]<rLowLnk[vState]) rLowLnk[vState]=rLowLnk[*sit];
       // Break recursion on find first
       if(rFilter.mMode & SccFilter::FindFirst)
       if(!rSccList.empty()) return;
    }
    // Successors that are not on the todo list ...
    else {
      // ... if they have a lower deph-first level ...
      if(rDfn[*sit]<rDfn[vState])  
      // ... and if the are on the stack at all ...
      if(rStackStates.Exists(*sit)) 
      // ... set the current low-link to the new minimum 
      if(rDfn[*sit]<rLowLnk[vState]) rLowLnk[vState]=rDfn[*sit];
    }
  }//end for: iterate successors

  // If the current state is the root of a SCC, record the result
  if((rLowLnk[vState]==rDfn[vState])) {
    FD_DF("SearchScc: retrieving SCC from stack, root " << vState);        
    bool record=true;
    // Technically, each state that fails to be part of a cycle forms an
    // SCC. We optionally ignore those trivial SCCs. They are identified by consisting of
    // one state only and having no (relevant) selfloop.
    if(rFilter.mMode & SccFilter::IgnoreTrivial) {
      if(rStack.top()==vState) {
        TransSet::Iterator tit = rGen.TransRelBegin(vState);
        TransSet::Iterator tit_end = rGen.TransRelEnd(vState);
        for(; tit != tit_end; ++tit) {
          // no self loop
          if(tit->X2!=vState) continue;
          // fail for avoid event
          if(rFilter.mMode & SccFilter::EventsAvoid)
	  if(rFilter.pEventsAvoid->Exists(tit->Ev)) 
            continue;
          /*
          // fail no required event
          if(rFilter.mMode & SccFilter::EventsRequire)
	  if(!rFilter.pEventsRequire->Exists(tit->Ev)) 
            continue;
	  */
          // found relevant selfloop
          break;
        }
        if(tit==tit_end) record=false;
      }
#ifdef FAUDES_DEBUG_FUNCTION
     if(!record) FD_DF("SearchScc: trivial loop, ignored");
#endif
    }
    /*
    // Inspect for required events (TODO: perhaps record path incl. events?)
    if(rFilter.mMode & SccFilter::EventsRequire) {
      bool reqevent=false;
      for(;vsit!=rStack.end();) {
        Idx x1, x2;
        x1=*vsit;
        if(++vsit==rStack.end()) x2=vState;
        else x2=*vsit;
        TransSet::Iterator tit = rGen.TransRelBegin(x1);
        TransSet::Iterator tit_end = rGen.TransRelEnd(x1);
        for(; tit != tit_end; ++tit) {
          // not on my path
          if(tit->X2!=x2) continue;
          // event not found
          if(!rFilter.pEventsRequire->Exists(tit->Ev)) continue;
          // found
          reqevent=true;
        }
      }
      if(reqevent==false) record=false;
    }
    */
    // Retrieve SCC from stack, inspect required states
    rSccList.push_back(StateSet());
    rRoots.Insert(vState);
    Idx top;
    bool reqstate=false;
    do {
      // Pop top of stack
      top=rStack.top();
      rStackStates.Erase(top);
      rStack.pop();
      // Record/track state requirements
      if(record) {
        rSccList.back().Insert(top);
        // Required state?
        if(rFilter.mMode & SccFilter::StatesRequire)
	if(rFilter.pStatesRequire->Exists(top)) 
          reqstate=true;
      }
    } while(top!=vState);
    // Invalidate missed requirements
    if(rFilter.mMode & SccFilter::StatesRequire)
    if(!reqstate)
      record=false;
    // Invalidate result
    if(!record) {
      rSccList.pop_back();
      rRoots.Erase(vState);
    }
  } // end if: record
}


// ComputeScc(Generator, SccList, Roots)
bool ComputeScc(
  const Generator& rGen,
  const SccFilter& rFilter,
  std::list<StateSet>& rSccList,
  StateSet& rRoots)
{
  FD_DF("CompteScc(" << rGen.Name() << ")");

  // inititalize results:
  rRoots.Clear();
  rSccList.clear();

  // initialize counter for depthfirstnumbers(DFN):
  int count=1;

  // provide local variables 
  std::stack<Idx> stack;
  StateSet stackstates;
  StateSet todostates;
  std::map<const Idx, int> dfn;
  std::map<const Idx, int> lowlnk;

  // initialise todo list
  if(rFilter.Mode() & SccFilter::IgnoreUnaccessible)
    todostates=rGen.AccessibleSet();
  else
    todostates=rGen.States();

  // start recursive depth-first search for Scc's:
  while(!todostates.Empty()) {
    SearchScc(*todostates.Begin(), count, rGen, rFilter, todostates, stack, stackstates, 
      dfn, lowlnk,rSccList, rRoots);
  }

  // done
  return !rSccList.empty();
}

// ComputeScc(Generator, SccList, Roots)
bool ComputeScc(
  const Generator& rGen,
  std::list<StateSet>& rSccList,
  StateSet& rRoots)
{
  FD_DF("CompteScc(" << rGen.Name() << ") [std]");

  // dummy
  const static SccFilter msFilter = SccFilter();

  // doit
  return ComputeScc(rGen,msFilter,rSccList,rRoots);
}


// HasScc(Generator, Scc)
bool HasScc(
  const Generator& rGen,
  const SccFilter& rFilter,
  StateSet& rScc)
{
  FD_DF("HasScc(" << rGen.Name() << ")");

  // inititalize result
  rScc.Clear();

  // initialize counter for depthfirstnumbers:
  int count=1;

  // provide local variables 
  std::stack<Idx> stack;
  StateSet stackstates;
  StateSet todostates;
  std::map<const Idx, int> dfn;
  std::map<const Idx, int> lowlnk;
  std::list<StateSet> scclist;
  StateSet roots;

  // initialise todo list
  if(rFilter.Mode() & SccFilter::IgnoreUnaccessible)
    todostates=rGen.AccessibleSet();
  else
    todostates=rGen.States();

  // start recursive depth-first search for Scc's:
  while(!todostates.Empty()) {
    SearchScc(*todostates.Begin(), count, rGen, rFilter, todostates, stack, stackstates, 
      dfn, lowlnk, scclist, roots);
  }

  // copy (!) result
  if(!scclist.empty()) rScc=*scclist.begin();

  // done
  return !rScc.Empty();
}

// HasScc(Generator, Scc)
bool HasScc(
  const Generator& rGen,
  const SccFilter& rFilter)
{
  FD_DF("HasScc(" << rGen.Name() << ") [boolean only]");

  // dummy
  StateSet scc;

  // doit
  return HasScc(rGen,rFilter,scc);
}


/*

// WriteStateSets(rStateSets)
// Write set of StateSet's to console.
void WriteStateSets(
  const std::set<StateSet>& rStateSets)
{
  FD_DF("WriteStateSets()");
  std::cout<<std::endl;
  if(rStateSets.empty()) {
    std::cout<<"WriteStateSets: Set of state sets is empty."<<std::endl;
    FD_DF("WriteStateSets: Set of state sets is empty.");
    return;
  }
  std::cout<<">WriteStateSets: Set of state sets begin:"<< std::endl;
  std::set<StateSet>::iterator StateSetit = rStateSets.begin();
  std::set<StateSet>::iterator StateSetit_end = rStateSets.end();
  for (; StateSetit != StateSetit_end; ++StateSetit) {
    std::cout<<std::endl<<"  >> State set begin:"<< std::endl;
    StateSet::Iterator sit = StateSetit->Begin();
    StateSet::Iterator sit_end = StateSetit->End();
    for (; sit != sit_end; ++sit) {
      std::cout<<"  >> "<<*sit<<std::endl;
    }
    std::cout<<"  >> State set end."<< std::endl;
  }  
  std::cout<<std::endl<<">WriteStateSets: Set of state sets end."<<std::endl;
}

// WriteStateSets(rGen,rStateSets)
// Write set of StateSet's to console.
// Use rGen for symbolic state names
void WriteStateSets(
  const Generator& rGen,
  const std::set<StateSet>& rStateSets)
{
  FD_DF("WriteStateSets()");
  std::cout<<std::endl;
  if(rStateSets.empty()) {
    std::cout<<"WriteStateSets: Set of state sets is empty."<<std::endl;
    FD_DF("WriteStateSets: Set of state sets is empty.");
    return;
  }
  std::cout<<">WriteStateSets: Set of state sets begin:"<< std::endl;
  std::set<StateSet>::iterator StateSetit = rStateSets.begin();
  std::set<StateSet>::iterator StateSetit_end = rStateSets.end();
  for (; StateSetit != StateSetit_end; ++StateSetit) {
    std::cout<<std::endl<<"  >> State set begin:"<< std::endl;
    std::cout<<"  "<<rGen.StateSetToString(*StateSetit)<<std::endl;
    std::cout<<"  >> State set end."<< std::endl;
  }  
  std::cout<<std::endl<<">WriteStateSets: Set of state sets end."<<std::endl;
}

*/

}//namespace faudes

