/** @file cfl_project.cpp projection and subset construction */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "cfl_project.h"

namespace faudes {

// UniqueInit(rGen&)
void UniqueInit(Generator& rGen) {
  Idx inituni;
  StateSet::Iterator lit;
  TransSet::Iterator tit;
  // check number of initial states
  if(rGen.InitStatesSize() == 1) return;
  // introduce new initial state
  if(rGen.StateNamesEnabled()) {
    std::string initname=rGen.UniqueStateName("InitUni");
    inituni = rGen.InsState(initname); 
  } else {
    inituni = rGen.InsState();
  }
  FD_DF("UniqueInit: introducing new initial state: " << inituni);
  // introduce outgoing transitions from initial state
  FD_DF("UniqueInit: introduce outgoing transitions: ");
  for (lit = rGen.InitStatesBegin(); lit != rGen.InitStatesEnd(); ++lit) {
    for (tit = rGen.TransRelBegin(*lit); tit != rGen.TransRelEnd(*lit); ++tit) {
      rGen.SetTransition(inituni, tit->Ev, tit->X2);
      FD_DF("UniqueInit:   " << inituni << "-" << tit->Ev << "-" << tit->X2);
    }
  }
  // mark the new init state if there exists an originally marked init state (tm 20101206)
  if(!(rGen.InitStates() *  rGen.MarkedStates()).Empty()){
    rGen.SetMarkedState(inituni);
    FD_DF("UniqueInit: set marked state: " << inituni);
  }
  // delete old istates
  rGen.ClearInitStates();
  // set inituni as new initial state	
  rGen.SetInitState(inituni);
}

// UniqueInit(rGen&,rResGen&)
void UniqueInit(const Generator& rGen, Generator& rResGen) {
  rResGen.Assign(rGen);
  UniqueInit(rResGen);
}

// Deterministic(rGen&, rResGen&)
void Deterministic(const Generator& rGen, Generator& rResGen) {
  // temporary vectors
  std::vector<StateSet> power_states;
  std::vector<Idx> det_states;
  Deterministic(rGen, power_states, det_states, rResGen);
}


// aDeterministic(rGen&, rResGen&)
void aDeterministic(const Generator& rGen, Generator& rResGen) {
  // prepare result to keep original alphabet
  Generator* pResGen = &rResGen;
  if(&rResGen==&rGen) {
    pResGen= rResGen.New();
  }
  // perform op
  Deterministic(rGen,*pResGen);
  // set old attributes
  pResGen->EventAttributes(rGen.Alphabet());
  // copy result
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }
}


// Deterministic(rGen&, rEntryStatesMap&, rResGen&)
void Deterministic(const Generator& rGen, std::map<Idx,StateSet>& rEntryStatesMap,
		   Generator& rResGen) {
  // prepare result:
  rEntryStatesMap.clear();
  // helpers:
  std::vector<StateSet> power_states;
  std::vector<Idx> det_states;
  // call Deterministic function
  Deterministic(rGen, power_states, det_states, rResGen);
  // build entry states map
  std::vector<StateSet>::size_type i;
  for (i = 0; i < power_states.size(); ++i) {
    rEntryStatesMap.insert(std::pair<Idx,StateSet>(det_states[i], power_states[i]));
  }
}


void Deterministic(const Generator& rGen, std::vector<StateSet>& rPowerStates,
		   std::vector<Idx>& rDetStates, Generator& rResGen) {
  // FIXME: implement the hashtable as a map<Idx, list< vector<int>::size_type > > 
  // or something like that

  // note: there is a demonstrative description of the multiway merge
  // algorithm in the master thesis. (Bernd Opitz)

  FD_DF("Deterministic(): core function #" << rGen.Size());

  // use pointer pResGen to result rResGen
  Generator* pResGen = &rResGen;
  if(&rResGen== &rGen) {
    pResGen= rResGen.New();
  }

  // prepare result
  pResGen->Clear();  
  rPowerStates.clear();
  rDetStates.clear();
  // set the name
  pResGen->Name(CollapsString("Det(" + rGen.Name() + ")"));
  // copy alphabet
  FD_DF("Deterministic A " << rGen.Alphabet().ToString());
  pResGen->InjectAlphabet(rGen.Alphabet());
  FD_DF("Deterministic B");

  // helpers
  TransSetEvX1X2 trel_evx1x2;
  rGen.TransRel(trel_evx1x2);
  typedef std::multimap< Idx,std::vector<StateSet>::size_type > T_HASHMAP;
  T_HASHMAP hashmap;
  std::vector<StateSet>::size_type current_vecindex;
  std::pair< std::map<StateSet,Idx>::iterator,bool > result;
  TransSet::Iterator transrel_end = rGen.TransRelEnd();
  StateSet newset;
  StateSet::Iterator lit;
  const Idx max_idx = std::numeric_limits<Idx>::max();
  
  // bail out on empty input 
  if(rGen.InitStatesEmpty()) {
    if(pResGen != &rResGen) {
      pResGen->Move(rResGen);
      delete pResGen;
    }
    FD_DF("Deterministic(): done (empty)"); 
    return;
  }
  
  Idx newstate = pResGen->InsInitState();
  // initialize rPowerStates with subset of initial states
  for (lit = rGen.InitStatesBegin(); lit != rGen.InitStatesEnd(); ++lit) {
    // clear set and insert single state
    newset.Insert(*lit);
    // if marked state set in res generator
    if (rGen.ExistsMarkedState(*lit)) {
      pResGen->SetMarkedState(newstate);
      FD_DF("Deterministic: setting as mstate: " << rGen.SStr(newstate));
    }
  }
  FD_DF("Deterministic: created subset of initial states {"
	<< newset.ToString() << "} with deterministic state index "
	<< rGen.SStr(newstate));
  // insert newset in rPowerStates
  rPowerStates.push_back(newset);
  rDetStates.push_back(newstate);
  hashmap.insert(std::make_pair(newset.Signature(), (Idx)rPowerStates.size() - 1));

  // iteration over all states
  for (current_vecindex = 0; current_vecindex < rPowerStates.size(); 
       ++current_vecindex) {
    FD_WPC(current_vecindex,rPowerStates.size(), "Deterministic(): current/size: "<<  current_vecindex << " / " << rPowerStates.size());
    FD_DF("Deterministic: current power set: {" 
	  << rPowerStates[current_vecindex].ToString() << "} -> " 
	  << rDetStates[current_vecindex]);

    std::vector<StateSet> newset_vec;
    std::vector<Idx> event_vec;

    // multiway merge begin
    FD_DF("Deterministic: starting multiway merge...");
    std::list<TransSet::Iterator> merge_iterators;
    std::vector<Transition> trans_vec;

    // add transset iterator at begin of each state's transitions
    TransSet::Iterator tit;
    for (lit = rPowerStates[current_vecindex].Begin(); 
	 lit != rPowerStates[current_vecindex].End(); ++lit) {
      tit = rGen.TransRelBegin(*lit);
      if (tit != rGen.TransRelEnd(*lit)) {
	merge_iterators.push_back(tit);
	FD_DF("Deterministic: added merge iterator: " << rGen.SStr(tit->X1) 
	      << "-" << rGen.EStr(tit->Ev) << "-" << rGen.SStr(tit->X2));
      }
    }

    // find first iterator with lowest event
    while (! merge_iterators.empty()) {
      Idx currentevent = max_idx;
      std::list<TransSet::Iterator>::iterator i; 
      std::list<TransSet::Iterator>::iterator currentit = merge_iterators.end();
      for (i = merge_iterators.begin(); i != merge_iterators.end(); ++i) {
	if ((*i)->Ev < currentevent) {
	  currentevent = (*i)->Ev;
	  currentit = i;
	}
      }
      // currentit now holds the iterator
      // currentevent holds the lowest event (lowest Idx)

      // merge all transitions with currentevent at each iterator in a row
      // this is a modification of multiway merge as after projection the
      // automaton most likely holds states with many transitions that share 
      // the same event. only merging the lowest transition and continue with
      // search for the lowest event again would be to slow here (because
      // of too much iterator dereferencing).
      Idx currentstate;
      while (currentit != merge_iterators.end()) {
	currentstate = (*currentit)->X1;
	TransSet::Iterator& j = *currentit;
	while (1) {
	  // remove iterator if it reaches the end of the transition set
	  if (j == transrel_end) {
	    std::list<TransSet::Iterator>::iterator tmpit = currentit;
	    ++currentit;
	    merge_iterators.erase(tmpit);
	    break;
	  }
	  // if current iterator is in its original state
	  else if (j->X1 == currentstate) {
	    // if the event is still the same add the transition
	    if (j->Ev == currentevent) {
	      trans_vec.push_back(*j); 
	      FD_DF("Deterine: adding transition to list: " 
		    << rGen.SStr(j->X1) << "-" << rGen.EStr(j->Ev) << "-"
		    << rGen.SStr(j->X2));
	    }
	    // else go to next iterator
	    else {
	      ++currentit;
	      break;
	    }
	  }
	  // if the iterator is beyond its original state remove it
	  else {
	    std::list<TransSet::Iterator>::iterator tmpit = currentit;
	    ++currentit;
	    merge_iterators.erase(tmpit);
	    break;
	  }
	  ++j;
	}
      }
    }

    // partition transition vector by events. optimizable? 
    FD_DF("Deterministic: partitioning the transition vector...");
    std::vector<Transition>::iterator tv_it;
    StateSet newset;
    Idx lastevent = 0;
    for (tv_it = trans_vec.begin(); tv_it != trans_vec.end(); ++tv_it) {
      if ((tv_it->Ev == lastevent) || (lastevent == 0)) {
	newset.Insert(tv_it->X2);
	lastevent = tv_it->Ev;
      }
      else {
	FD_DF("Deterministic: partition: {" << newset.ToString() 
	      << "} with event " << rGen.EStr(lastevent));
	newset_vec.push_back(newset);
	event_vec.push_back(lastevent);
	newset.Clear();
	newset.Insert(tv_it->X2);
	lastevent = tv_it->Ev;
      }
    }
    if (! newset.Empty()) {
      FD_DF("Deterministic: partition: {" << newset.ToString() 
	    << "} with event " << rGen.EStr(lastevent));
      newset_vec.push_back(newset);
      event_vec.push_back(lastevent);
    }
    FD_DF("Deterministic: partitioning the transition vector finished");
    FD_DF("Deterministic: multiway merge finished");
    // multiway merge end

    std::vector<StateSet>::size_type nsv_index;
    for (nsv_index = 0; nsv_index < newset_vec.size(); ++nsv_index) {
      StateSet& currentset = newset_vec[nsv_index];
      Idx currentevent = event_vec[nsv_index];
      Idx tmp_x2 = 0;
      Idx sig = currentset.Signature();
      // test if newset signature is already known
      std::pair<T_HASHMAP::iterator,T_HASHMAP::iterator> phit
	= hashmap.equal_range(sig);
      T_HASHMAP::iterator hit = phit.first;
      for (hit = phit.first; hit != phit.second; ++hit) {
	// test set of every matching signature for equality
	if (currentset == rPowerStates[hit->second]) {
	  tmp_x2 = rDetStates[hit->second];
	  break;
	}
      }

      // if new set is unique within the existing power sets
      if (tmp_x2 == 0) {
	// create new state in res generator
	tmp_x2 = pResGen->InsState();
	// insert newset in rPowerStates and get iterator,bool pair
	rPowerStates.push_back(currentset);
	rDetStates.push_back(tmp_x2);
	hashmap.insert(std::make_pair(sig, (Idx)rPowerStates.size() - 1));
	FD_DF("Deterministic: added new state " 
	      << rGen.SStr(tmp_x2) 
	      << " for new subset {" <<	currentset.ToString() << "}");
	// set marked if one of the states in current set is marked
	for (lit = currentset.Begin(); lit != currentset.End(); ++lit) {
	  if (rGen.ExistsMarkedState(*lit)) {
	    pResGen->SetMarkedState(tmp_x2);
	    break;
	  }
	}
      }
      // introduce transition
      pResGen->SetTransition(rDetStates[current_vecindex], currentevent, tmp_x2);
    }
  }
  // fix names
  if (rGen.StateNamesEnabled() && pResGen->StateNamesEnabled()) {
    FD_DF("Deterministic: fixing names...");
    // rPowerStates / rDetStates index "iterator"
    std::vector<StateSet>::size_type i;
    // deterministic states iterator
    std::vector<Idx>::const_iterator dit;
    for (i = 0; i < rPowerStates.size(); ++i) {
      // temporary state name
      std::string name = "{";
      for (lit = rPowerStates[i].Begin();  lit != rPowerStates[i].End(); ++lit) {
	if (rGen.StateName(*lit) != "") {
	  name = name + rGen.StateName(*lit) + ",";
	}
	else {
	  name = name + ToStringInteger(*lit) + ",";
	}
      }
      name.erase(name.length() - 1);
      name = name + "}";
      FD_DF("Deterministic: setting state name \"" << name << "\" for index "
	    << rDetStates[i]);
      pResGen->StateName(rDetStates[i], name);
    }
  }
  
  // move pResGen to rResGen
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }

  FD_DF("Deterministic(): core function: done");
    
}



// ProjectNonDet(rGen&, rProjectAlphabet&)
// (version 2009/05, Barthel)
void ProjectNonDet(Generator& rGen, const EventSet& rProjectAlphabet) {

  // HELPERS:

  StateSet reach, reachext; // StateSet for reachable states
  StateSet todo; // states todo
  StateSet done; // states done 
  Idx currentstate; // the currently processed state
  StateSet::Iterator lit;
  StateSet::Iterator lit2;
  TransSet::Iterator tit;
  TransSet::Iterator tit_end;

  // NAME
  std::string name=CollapsString("ProjectNonDet(" + rGen.Name() + ")");

  // ALGORITHM:
  // set all locally reachable states to init states
  for (lit = rGen.InitStatesBegin(); lit != rGen.InitStatesEnd(); ++lit) {
    reach.Clear();
    LocalAccessibleReach(rGen, rProjectAlphabet, *lit, reach);
    for (lit2 = reach.Begin(); lit2 != reach.End(); ++lit2) {
      if (!rGen.InitStates().Exists(*lit2)) {
	FD_DF("ProjectNonDet: set initial state: " << rGen.SStr(*lit2));
	rGen.SetInitState(*lit2);
      }
    }
  }
  // initialize algorithm by adding init states to todo 
  for (lit = rGen.InitStatesBegin(); lit != rGen.InitStatesEnd(); ++lit) {
    FD_DF("ProjectNonDet: todo add: " << rGen.SStr(*lit));
    todo.Insert(*lit);
  }
    
  // process todo stack
  while (!todo.Empty()) {
    currentstate = *todo.Begin();
    todo.Erase(*todo.Begin());
    done.Insert(currentstate); // mark as done
    //std::cout << "###" << std::endl;
    FD_DF("ProjectNonDet: current state: " << rGen.SStr(currentstate));

    // comp accessible reach
    reach.Clear();
    LocalAccessibleReach(rGen, rProjectAlphabet, currentstate, reach);
    FD_DF("ProjectNonDet: local reach: " << reach.ToString());
		
    // relink outgoing transitions
    FD_DF("ProjectNonDet: relinking outgoing transitions...");
    for (lit = reach.Begin(); lit != reach.End(); ++lit) {
      tit = rGen.TransRelBegin(*lit);
      tit_end = rGen.TransRelEnd(*lit);
      for (; tit != tit_end; ++tit) {
	if (rProjectAlphabet.Exists(tit->Ev)) {
	  FD_DF("ProjectNonDet: relinking transition: " << rGen.TStr(*tit) << " to " << rGen.SStr(currentstate));
	  rGen.SetTransition(currentstate, tit->Ev, tit->X2);
	  //if (!done.Exists(tit->X2) && !todo.Exists(*lit2)) {
	  if (!done.Exists(tit->X2)) {
	    FD_DF("ProjectNonDet: todo insert: " << rGen.SStr(tit->X2));
	    todo.Insert(tit->X2);
	  }
	  // add transistions to all states in extended local reach
	  reachext.Clear();
	  LocalAccessibleReach(rGen, rProjectAlphabet, tit->X2, reachext);
	  FD_DF("ProjectNonDet: local reach from state " << tit->X2 << ": " << reachext.ToString());
	  for (lit2 = reachext.Begin(); lit2 != reachext.End(); ++lit2) {
	    if (!rGen.ExistsTransition(tit->X2, tit->Ev, *lit2) && (tit->X2 != *lit2)) {
	      rGen.SetTransition(tit->X1, tit->Ev, *lit2);
	      FD_DF("ProjectNonDet: setting transition: " << rGen.SStr(tit->X1) << "-" << rGen.EStr(tit->Ev) << "-" << rGen.SStr(*lit2));
	      //if (!done.Exists(*lit2) && !todo.Exists(*lit2)) {
	      if (!done.Exists(*lit2)) {
		FD_DF("ProjectNonDet: todo insert: " << rGen.SStr(tit->X2));
		todo.Insert(*lit2);
	      }
	    }
	  }
	  //std::cout << "---" << std::endl;
	}
      }
      // marked status test
      if (rGen.ExistsMarkedState(*lit)) {
	FD_DF("ProjectNonDet: setting marked state " << rGen.SStr(currentstate));
	rGen.SetMarkedState(currentstate);
      }
    }
	 
    // remove all transitions that leave current state 
    // with an invisible event
    tit = rGen.TransRelBegin(currentstate);
    tit_end = rGen.TransRelEnd(currentstate);
    while(tit != tit_end) {
      FD_DF("ProjectNonDet: current transition: " << rGen.SStr(tit->X1)
	    << "-" << rGen.EStr(tit->Ev) << "-" << rGen.SStr(tit->X2));
      if (! rProjectAlphabet.Exists(tit->Ev)) {
	FD_DF("ProjectNonDet: deleting current transition");
	TransSet::Iterator tit_tmp = tit;
	++tit;
	rGen.ClrTransition(tit_tmp);
      } else {
	++tit;
      }
    }

  }// todo loop

  // inject projection alphabet, keep Attributes
  rGen.RestrictAlphabet(rProjectAlphabet);

  // make accessile
  rGen.Accessible();
	
  // set name
  rGen.Name(name);
}

// wrapper
void aProjectNonDet(Generator& rGen, const EventSet& rProjectAlphabet) {
  ProjectNonDet(rGen,rProjectAlphabet);
}

/*

// reference, 2008/12

// Project(rGen&, rProjectAlphabet&)
void ProjectNonDet(Generator& rGen, const EventSet& rProjectAlphabet) {

// HELPERS:

StateSet reach; // StateSet for reachable states
std::stack<Idx> todo; // todo stack
StateSet done; // done set
Idx currentstate; // the currently processed state
StateSet::Iterator lit; 
TransSet::Iterator tit;
TransSet::Iterator tit_end;

// ALGORITHM:
// initialize algorithm by pushing init states on todo stack
for (lit = rGen.InitStatesBegin(); lit != rGen.InitStatesEnd(); ++lit) {
FD_DF("ProjectNonDet: todo add: " << rGen.SStr(*lit));
todo.push(*lit);
}

// process todo stack
while (! todo.empty()) {
currentstate = todo.top();
todo.pop();
done.Insert(currentstate); // mark as done
FD_DF("ProjectNonDet: current state: " << rGen.SStr(currentstate));

// comp accessible reach
reach.Clear();
LocalAccessibleReach(rGen, rProjectAlphabet, currentstate, reach);
FD_DF("ProjectNonDet: local reach: " << reach.ToString());

// remove all transitions that leave current state 
// with an invisible event
tit = rGen.TransRelBegin(currentstate);
tit_end = rGen.TransRelEnd(currentstate);
while(tit != tit_end) {
FD_DF("ProjectNonDet: current transition: " << rGen.SStr(tit->X1)
<< "-" << rGen.EStr(tit->Ev) << "-" << rGen.SStr(tit->X2));
if (! rProjectAlphabet.Exists(tit->Ev)) {
FD_DF("ProjectNonDet: deleting current transition");
TransSet::Iterator tit_tmp = tit;
++tit;
rGen.ClrTransition(tit_tmp);
} else {
++tit;
}
}
		
// relink outgoing transitions
FD_DF("ProjectNonDet: relinking outgoing transitions...");
for (lit = reach.Begin(); lit != reach.End(); ++lit) {
tit = rGen.TransRelBegin(*lit);
tit_end = rGen.TransRelEnd(*lit);
for (; tit != tit_end; ++tit) {
if (rProjectAlphabet.Exists(tit->Ev)) {
FD_DF("ProjectNonDet: relinking transition: " << rGen.TStr(*tit) << " to " << rGen.SStr(currentstate));
rGen.SetTransition(currentstate, tit->Ev, tit->X2);
if (! done.Exists(tit->X2)) {
FD_DF("ProjectNonDet: todo push: " << rGen.SStr(tit->X2));
todo.push(tit->X2);
}
}
}
// marked status test
if (rGen.ExistsMarkedState(*lit)) {
FD_DF("ProjectNonDet: setting marked state " << rGen.SStr(currentstate));
rGen.SetMarkedState(currentstate);
}
}
}

// inject projection alphabet
rGen.InjectAlphabet(rProjectAlphabet);
	
// set name
rGen.Name(CollapsString("Pro(" + rGen.Name() + ")"));
}

*/
			
// Project(rGen, rProjectAlphabet, rResGen&)
void Project(const Generator& rGen, const EventSet& rProjectAlphabet, Generator& rResGen) {
  FD_DF("Project(...)");
  // temporarily assign rGen to rResGen
  if(&rResGen != &rGen) rResGen.Assign(rGen);
  // turn of state names
  bool se= rResGen.StateNamesEnabled();
  rResGen.StateNamesEnabled(false);
  // project non det version
  ProjectNonDet(rResGen, rProjectAlphabet);
  // make deterministic
  Generator* g2 = rGen.New(); 
  g2->StateNamesEnabled(false);
  Deterministic(rResGen, *g2);
  // minimize states and write result to rResGen
  StateMin(*g2, rResGen);
  delete g2;
  // restore state names
  rResGen.StateNamesEnabled(se);
  // set name
  rResGen.Name("Project("+CollapsString(rGen.Name()+")")); 
}

	 
// Project(rGen, rProjectAlphabet, rResGen&)
void aProject(const Generator& rGen, const EventSet& rProjectAlphabet, Generator& rResGen) {
  // prepare result to keep original alphabet
  Generator* pResGen = &rResGen;
  if(&rResGen== &rGen) {
    pResGen= rResGen.New();
  }
  // perform op
  Project(rGen,rProjectAlphabet,*pResGen);
  // set old attributes
  pResGen->EventAttributes(rGen.Alphabet());
  // copy result
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }
}



// Project(rGen, rProjectAlphabet, rEntryStatesMap&, rResGen&)
void Project(const Generator& rGen, const EventSet& rProjectAlphabet,
	     std::map<Idx,StateSet>& rEntryStatesMap, Generator& rResGen) {
  FD_DF("Project(...)");
  // temporary entry state map
  std::map<Idx,StateSet> tmp_entrystatemap;
  // temporarily assign rGen to rResGen
  if(&rResGen != &rGen) rResGen.Assign(rGen); 
  // project tmp with respect to palphabet
  ProjectNonDet(rResGen, rProjectAlphabet); 
  // put deterministic result into tmp
  Generator* tmp = rGen.New();
  Deterministic(rResGen, tmp_entrystatemap, *tmp);
  // write entry state map for minimized generator
  std::vector<StateSet> subsets;
  std::vector<Idx> newindices;
  // minimize states and rewrite result to rResGen
  StateMin(*tmp, rResGen, subsets, newindices);
  // build entry state map
  std::vector<StateSet>::size_type i;
  std::map<Idx,StateSet>::iterator esmit;
  StateSet::Iterator sit;
  for (i = 0; i < subsets.size(); ++i) {	
    StateSet tmpstates;
    for (sit = subsets[i].Begin(); sit != subsets[i].End(); ++sit) {
      esmit = tmp_entrystatemap.find(*sit);
#ifdef FAUDES_DEBUG_CODE
      if (esmit == tmp_entrystatemap.end()) {
        FD_DF("project internal error");
	abort();
      }
#endif
      // insert entry states in temporary StateSet
      tmpstates.InsertSet(esmit->second);
    }

    rEntryStatesMap.insert(std::make_pair(newindices[i], tmpstates));
  }
  delete tmp;
}


// InvProject(rGen&, rProjectAlphabet)
void InvProject(Generator& rGen, const EventSet& rProjectAlphabet) {
  // test if the alphabet of the generator is included in the given alphabet
  if(! (rProjectAlphabet >= (EventSet) rGen.Alphabet() ) ){
    std::stringstream errstr;
    errstr << "Input alphabet has to contain alphabet of generator \"" << rGen.Name() << "\"";
    throw Exception("InvProject(Generator,EventSet)", errstr.str(), 506);
  }
  EventSet newevents = rProjectAlphabet - rGen.Alphabet();
  // insert events into generator
  rGen.InsEvents(newevents);
  FD_DF("InvProject: adding events \"" << newevents.ToString() 
	<< "\" at every state");
  StateSet::Iterator lit;
  EventSet::Iterator eit;
  for (lit = rGen.StatesBegin(); lit != rGen.StatesEnd(); ++lit) {
    LoopCallback(); // should only be an issue for very large generators 
    for (eit = newevents.Begin(); eit != newevents.End(); ++eit) {
      rGen.SetTransition(*lit, *eit, *lit);
    }
  }
}

// InvProject(rGen&, rProjectAlphabet)
void aInvProject(Generator& rGen, const EventSet& rProjectAlphabet) {
  FD_DF("aInvProject(..)");
  // see whether the generator can digest attribues
  if(!rGen.Alphabet().AttributeTry(rProjectAlphabet.Attribute())) {
    InvProject(rGen,rProjectAlphabet);
    return;
  }
  // record extra events
  EventSet newevents = rProjectAlphabet - rGen.Alphabet();
  // perform
  InvProject(rGen,rProjectAlphabet);
  // copy all attributes from input alphabets
  FD_DF("aInvProject(..): fixing attributes: source " << typeid(rProjectAlphabet.Attribute()).name() <<
	  " dest " << typeid(rGen.Alphabet().Attribute()).name());
  for(EventSet::Iterator eit=newevents.Begin(); eit!=newevents.End(); ++eit)  
    rGen.EventAttribute(*eit,rProjectAlphabet.Attribute(*eit));
}


// InvProject
void aInvProject(
  const Generator& rGen,
  const EventSet& rProjectAlphabet,
  Generator& rResGen)
{
  rResGen.Assign(rGen);
  aInvProject(rResGen, rProjectAlphabet);
}




// CreateEntryStatesMap(rRevEntryStatesMap, rEntryStatesMap)
void CreateEntryStatesMap(const std::map<StateSet,Idx>& rRevEntryStatesMap,
			  std::map<Idx,StateSet>& rEntryStatesMap) {
  std::map<StateSet,Idx>::const_iterator it;
  for (it = rRevEntryStatesMap.begin(); it != rRevEntryStatesMap.end(); ++it) {
    rEntryStatesMap.insert(std::make_pair(it->second, it->first));
  }
}



} // namespace faudes
